/*******************************************************************************
 *
 * blueMarine - open source photo workflow
 * =======================================
 *
 * Copyright (C) 2003-2009 by Fabrizio Giudici
 * Project home page: http://bluemarine.tidalwave.it
 *
 *******************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 *
 *******************************************************************************
 *
 * $Id$
 *
 ******************************************************************************/
package it.tidalwave.metadata.impl;

import javax.annotation.Nonnull;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import it.tidalwave.util.logging.Logger;
import org.openide.loaders.DataObject;
import org.openide.util.Lookup;
import it.tidalwave.swing.beansbinding.AbstractBean;
import it.tidalwave.metadata.Metadata;
import it.tidalwave.metadata.MetadataItemHolder;
import it.tidalwave.metadata.MetadataItemHolderList;
import it.tidalwave.metadata.NoSuchMetadataTypeException;
import it.tidalwave.metadata.spi.ClassMap;
import it.tidalwave.metadata.spi.MetadataItemInterceptor;
import it.tidalwave.metadata.spi.MetadataItemProvider;
import it.tidalwave.metadata.spi.MetadataSpiUtils;

/*******************************************************************************
 *
 * The default implementation of Metadata.
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public class MetadataImpl extends AbstractBean implements Metadata 
  {
    private static final String CLASS = MetadataImpl.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private static final String PROP_LATEST_MODIFICATION_TIME = "latestModificationTime";

    /** Instance counter. */
    private static int instanceCount;
    
    /** A map of MetadataItemProviders indexed by their product class. */
    protected static final ClassMap<MetadataItemProvider<?>> providerMapByClass = new ClassMap<MetadataItemProvider<?>>();
    
    /** The list of interceptors. */
    protected static final List<MetadataItemInterceptor<?>> interceptors = new ArrayList<MetadataItemInterceptor<?>>();
    
    /** The reference to the bound DataObject. */
    @Nonnull 
    private final DataObject dataObject;
   
    /** There is a map of MetadataItemHolders indexed by their class (note: in case of synthetically enhanced 
     *  items, the key is the original class, not the enhanced one, but this detail is handled by ClassMap). 
     *  This map does not use weak references on purpose, since e.g. during batch import items would be frequently 
     *  lost and they are reloaded multiple times from disk, causing a very bad performance. 
     *  The map is itself put into a map that is indexes by StorageType (e.g. EXTERNAL, INTERNAL). */
    protected final Map<StorageType, ClassMap<List<MetadataItemHolder<?>>>> holderMapByOption =
                new HashMap<StorageType, ClassMap<List<MetadataItemHolder<?>>>>();
    
    private Date oldLatestModificationTime;
    
    /***************************************************************************
     *
     * 
     * 
     **************************************************************************/
    static
      {
        for (final MetadataItemProvider metadataProvider : Lookup.getDefault().lookupAll(MetadataItemProvider.class))
          {
            providerMapByClass.put(metadataProvider.getItemClass(), metadataProvider);
          }
        
        for (final MetadataItemInterceptor interceptor : Lookup.getDefault().lookupAll(MetadataItemInterceptor.class))
          {
            interceptors.add(interceptor);
          }
        
        logger.info("MetadataItemProviders: %s", providerMapByClass);
        logger.info("MetadataItemInterceptors: %s", interceptors);
      }
    
    /***************************************************************************
     *
     * 
     * 
     **************************************************************************/
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("SIC_INNER_SHOULD_BE_STATIC_ANON")
    private final PropertyChangeListener propertyListener = new PropertyChangeListener() 
      {
        @Override
        public void propertyChange (@Nonnull final PropertyChangeEvent event) 
          {
            logger.finer("propertyChange(%s)", event);
//            if (event.getPropertyName().endsWith("Available")) // TODO
//              {
//                firePropertyChange("available", oldAvailable, isAvailable());
//              }
//
            // FIXME: this is disabled since it gets called from the setting of beans
            // Probably you should not bind beans to listener BEFORE populating them for the first time.
            // when they are read from the database and trigger spurious operations.
//            if (event.getPropertyName().equals(PROP_LATEST_MODIFICATION_TIME))
//              {
//                final Date latestModificationTime = getLatestModificationTime();
//                firePropertyChange(PROP_LATEST_MODIFICATION_TIME, oldLatestModificationTime, latestModificationTime);
//                oldLatestModificationTime = latestModificationTime;
//              }
          }
      };
      
    /***************************************************************************
     *
     * 
     * 
     **************************************************************************/
    protected MetadataImpl (@Nonnull final DataObject dataObject)
      {
        logger.fine("MetadataImpl(%s)", dataObject);
            
        if (dataObject == null)
          {
            throw new IllegalArgumentException("null dataObject");
          }
        
        this.dataObject = dataObject;
        instanceCount++;
        logger.finest("Total number of Metadata instances: %d", instanceCount);
        // don't call getLatestModificationTime() now or it will load all items
//        oldLatestModificationTime = getLatestModificationTime();
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Override
    public boolean isItemAvailable (@Nonnull final Class<?> metadataItemClass,
                                    @Nonnull final FindOption ... options)
      throws NoSuchMetadataTypeException
      {
        final StorageType storageType = MetadataSpiUtils.findOption(options, StorageType.DEFAULT, StorageType.class);
        final ClassMap<List<MetadataItemHolder<?>>> holderMapByClass = findHolderMap(storageType);
        return holderMapByClass.containsKey(metadataItemClass);
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Nonnull 
    @Override
    public synchronized <Item> MetadataItemHolder<Item> findOrCreateItem (@Nonnull final Class<Item> metadataItemClass, 
                                                                          @Nonnull final FindOption ... options)
      throws NoSuchMetadataTypeException 
      {
        logger.fine("findOrCreateItem(%s, %s) - %s", metadataItemClass, Arrays.toString(options), dataObject);
        final List<MetadataItemHolder<Item>> items = findOrCreateItems(metadataItemClass, options);
        return items.get(0); // is empty for classes not supported
      }
                
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Nonnull 
    @Override
    public synchronized <Item> MetadataItemHolderList<Item> findOrCreateItems (@Nonnull final Class<Item> metadataItemClass,
                                                                               @Nonnull final FindOption ... options)
      throws NoSuchMetadataTypeException
      {
        logger.fine("findOrCreateItems(%s, %s) - %s", metadataItemClass, Arrays.toString(options), dataObject);
            
        if (metadataItemClass == null)
          {
            throw new IllegalArgumentException("metadataItemClass is null");
          }
        
        final List<MetadataItemHolder<Item>> result = new ArrayList<MetadataItemHolder<Item>>();
        
        final StorageType storageType = MetadataSpiUtils.findOption(options, StorageType.DEFAULT, StorageType.class);
        final List<StorageType> storageTypes = new ArrayList<StorageType>();

        if (storageType != StorageType.ANY_TYPE)
          {
            storageTypes.add(storageType);
          }
        else
          {
            for (final StorageType type : StorageType.values())
              {
                if ((type != StorageType.ANY_TYPE) && (type != StorageType.NONE))
                  {
                    storageTypes.add(type);
                  }
              }
          }

        for (final StorageType type : storageTypes)
          {
            // FIXME: use Arguments and introduce options2 = Arguments.replaced(StorageType.ANY_TYPE, type);
            final FindOption[] options2 = options.clone();

            for (int i = 0; i < options2.length; i++)
              {
                if (options2[i] == StorageType.ANY_TYPE)
                  {
                    options2[i] = type;
                    break;
                  }
              }
 
            final ClassMap<List<MetadataItemHolder<?>>> holderMapByClass = findHolderMap(type);
            List<MetadataItemHolder<?>> holders = holderMapByClass.get(metadataItemClass);

            if (holders == null)
              {
                final MetadataItemProvider<Item> metadataProvider = (MetadataItemProvider<Item>)providerMapByClass.get(metadataItemClass);

                if (metadataProvider == null)
                  {
                    throw new NoSuchMetadataTypeException(metadataItemClass);
                  }

                holders = new ArrayList<MetadataItemHolder<?>>();

                for (final MetadataItemHolder<Item> holder : metadataProvider.findOrCreateItems(dataObject, options2))
                  {
                    fireNotifyItemCreated(holder);
                    holder.addPropertyChangeListener(PROP_LATEST_MODIFICATION_TIME, propertyListener);
                    holders.add(holder);
                  }

                holderMapByClass.put(metadataItemClass, holders);
    //                metadataItemMapByClass.put(metadataItemClass, new WeakReference<List<Object>>(items));
              }

            for (final Object holder : holders)
              {
                result.add((MetadataItemHolder<Item>)holder);
              }
           }
        
        logger.finer(">>>> returning %s", result);
        return new MetadataItemHolderList<Item>(result);
      }
                
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Override
    public <Item> int storeItem (@Nonnull final MetadataItemHolder<Item> holder, @Nonnull final StoreOption ... options)
      throws NoSuchMetadataTypeException
      {
        logger.info("storeItem(%s, %s)", holder, Arrays.toString(options));
        
        if (holder == null)
          {
            throw new IllegalArgumentException("holder is mandatory");
          }
        
        MetadataSpiUtils.validateOptions(options);
        
        final Class<Item> itemClass = (Class<Item>)holder.get().getClass();
        final MetadataItemProvider<Item> metadataProvider = (MetadataItemProvider<Item>)providerMapByClass.get(itemClass);

        if (metadataProvider == null)
          {
            throw new NoSuchMetadataTypeException(itemClass);
          }
        
        return metadataProvider.storeItem(getDataObject(), holder, options);
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Nonnull 
    @Override
    public Set<Class<?>> getItemClasses() 
      {
        return new CopyOnWriteArraySet<Class<?>>(providerMapByClass.keySet());
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Nonnull 
    @Override
    public Date getLatestModificationTime()  
      {
        logger.fine("getLatestModificationTime()");
        Date result = null;
        
        // Force loading all items, or you won't be able to compute the most recent date.
        // FIXME: I don't like this, but when we will have a persistent model for Metadata, the
        // value could be just loaded from the persistent layer, and recomputed here only
        // when we need an update.
        for (final MetadataItemProvider<?> provider : providerMapByClass.values())
          {
            try 
              {
                findOrCreateItem(provider.getItemClass(), StorageType.ANY_TYPE);
              } 
            catch (NoSuchMetadataTypeException e) 
              {
                logger.warning("Unexpected " + e); // should never happen here
              }
          }

        for (final StorageType type : StorageType.values())
          {
            if (type != StorageType.ANY_TYPE)
              {
                final ClassMap<List<MetadataItemHolder<?>>> holderMapByClass = findHolderMap(type);

                for (final Entry<Class<?>, List<MetadataItemHolder<?>>> entry : holderMapByClass.entrySet())
                  {
                    final List<MetadataItemHolder<?>> holders = entry.getValue();
                    final MetadataItemHolder<?> holder = holders.get(0);
                    final Date date = holder.getLatestModificationTime();
//                    logger.finest(">>>> holder %s, latestModificationTime: %s", holder, date));

                    if ((result == null) || ((date != null) && date.after(result)))
                      {
                        result = date;
                      }
                  }
              }
          }
        
        return result;
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Nonnull 
    @Override
    public DataObject getDataObject()
      {
        return dataObject;
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Override
    @Nonnull 
    final public String toString()
      {
        return String.format("MetadataImpl[%s]", dataObject);    
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     * 
     **************************************************************************/
    @Override
    protected void finalize() 
      throws Throwable 
      {
        instanceCount--;
        logger.finest("Total number of Metadata instances: %d", instanceCount);
        super.finalize();
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    @Nonnull 
    private ClassMap<List<MetadataItemHolder<?>>> findHolderMap (@Nonnull final StorageType storageType)
      {
        // Do not clone results, should return the live pointers
        ClassMap<List<MetadataItemHolder<?>>> result = null;

        if (storageType == StorageType.ANY_TYPE)
          {
            throw new IllegalArgumentException("storageType == ANY_TYPE, but it's not a real key");
          }

        result = holderMapByOption.get(storageType);

        if (result == null)
          {
            result = new ClassMap<List<MetadataItemHolder<?>>>();
            holderMapByOption.put(storageType, result);
          }
        
        return result;
      }
    
    /***************************************************************************
     *
     * Postprocesses a newly created metadata item.
     * 
     * @param  holder  the item to postprocess 
     * 
     **************************************************************************/
    private <T> void fireNotifyItemCreated (@Nonnull final MetadataItemHolder<T> holder)
      {
        logger.fine("fireNotifyItemCreated(%s)", holder);

        final Class<T> itemClass = (Class<T>)holder.get().getClass();
        
        for (final MetadataItemInterceptor<?> interceptor : interceptors)
          {
            try 
              {
                if (interceptor.getItemClass().isAssignableFrom(itemClass))
                  {
                    ((MetadataItemInterceptor<T>)interceptor).notifyItemCreated(this, holder);
                  }
              }
            catch (Throwable t)
              {
                logger.throwing(CLASS, "fireNotifyItemCreated()", t);
                logger.warning("fireNotifyItemCreated() threw exception: %s", interceptor);
              }
          }
      }
  }
